from typing import Optional, Type

from pydantic import BaseModel, Field, create_model


def make_optional(model: Type[BaseModel]) -> Type[BaseModel]:
    """Make all fields of a Pydantic model optional while preserving validation constraints."""
    fields = {}

    model_schema = model.model_json_schema()
    properties = model_schema.get("properties", {})

    for name, field in model.model_fields.items():

        field_schema = properties.get(name, {})

        constraints = {}

        if "minimum" in field_schema:
            constraints["ge"] = field_schema["minimum"]
        if "maximum" in field_schema:
            constraints["le"] = field_schema["maximum"]
        if "exclusiveMinimum" in field_schema:
            constraints["gt"] = field_schema["exclusiveMinimum"]
        if "exclusiveMaximum" in field_schema:
            constraints["lt"] = field_schema["exclusiveMaximum"]
        if "minLength" in field_schema:
            constraints["min_length"] = field_schema["minLength"]
        if "maxLength" in field_schema:
            constraints["max_length"] = field_schema["maxLength"]
        if "pattern" in field_schema:
            constraints["pattern"] = field_schema["pattern"]
        if "description" in field_schema:
            constraints["description"] = field_schema["description"]

        if "error_messages" in field_schema:
            constraints["error_messages"] = field_schema["error_messages"]

        fields[name] = (Optional[field.annotation], Field(default=None, **constraints))

    return create_model(f"Optional{model.__name__}", **fields, __base__=model)



from typing import Any, List, Optional, Protocol

from ..schemas import FindUniqueValues


class RepositoryProtocol(Protocol):
    def read_by_options(
        self, schema: Any, searchable_fields: Optional[List[str]] = None
    ) -> Any: ...

    def read_by_id(self, id: int) -> Any: ...

    def create(self, schema: Any) -> Any: ...

    def bulk_create(self, schemas: List[Any]) -> Any: ...

    def update(self, id: int, schema: Any) -> Any: ...

    def update_attr(self, id: int, attr: str, value: Any) -> Any: ...

    def whole_update(self, id: int, schema: Any) -> Any: ...

    def delete_by_id(self, id: int) -> Any: ...

    def get_unique_values(self, schema: Any) -> Any: ...

    def close_scoped_session(self): ...

    def read_by_attr(self, attr: str, value: Any) -> Any: ...
class BaseService:
    def __init__(self, repository: RepositoryProtocol) -> None:
        self._repository = repository

    def get_list(
        self, schema: Any, searchable_fields: Optional[List[str]] = None
    ) -> Any:
        return self._repository.read_by_options(schema, searchable_fields, eager=True)

    def get_by_id(self, id: int) -> Any:
        return self._repository.read_by_id(id, eager=True)

    def add(self, schema: Any) -> Any:
        return self._repository.create(schema)
    
    def bulk_add(self, schemas: List[Any]) -> Any:
        return self._repository.bulk_create(schemas)

    def patch(self, id: int, schema: Any) -> Any:
        return self._repository.update(id, schema)

    def patch_attr(self, id: int, attr: str, value: Any) -> Any:
        return self._repository.update_attr(id, attr, value)

    def put_update(self, id: int, schema: Any) -> Any:
        return self._repository.whole_update(id, schema)

    def remove_by_id(self, id: int) -> Any:
        return self._repository.delete_by_id(id)

    def get_unique_values(self, schema: FindUniqueValues) -> dict:
        return self._repository.get_unique_values(schema)
    
    def get_by_attr(self, attr: str, value: Any) -> Any:
        return self._repository.read_by_attr(attr, value)

    def close_scoped_session(self):
        self._repository.close_scoped_session()



from .base_schema import FindBase, ModelBaseInfo, FindResult, SearchOptions, FindUniqueValues, UniqueValuesResult, FilterSchema, SortOrder
from ..utils.schema import make_optional

__all__ = ["FindBase", "ModelBaseInfo", "FindResult", "SearchOptions", "make_optional", "FindUniqueValues", "UniqueValuesResult", "FilterSchema", "SortOrder"]